from lolpop.component.base_component import BaseComponent

class BaseModelRepository(BaseComponent): 

    def register_model(self, model, *args, **kwargs): 
        pass 

    def promote_model(self, model, *args, **kwargs): 
        pass 

    def approve_model(self, model, *args, **kwargs): 
        pass 