from lolpop.component.base_component import BaseComponent

class BaseDataProfiler(BaseComponent): 

    def profile_data(self, data, *args, **kwargs): 
        pass 

    def compare_data(self, data, prev_data, *args, **kwargs): 
        pass
